@@ -12,8 +12,11 @@ use clap::clap_app;
 use rand::prelude::*;
 
 const VERSION: &str = env!("CARGO_PKG_VERSION");
+const TARGET_DIR: &str = "target/flowistry";
 
 fn main() {
+  env_logger::init();
+
   let flowistry_rustc_path = std::env::current_exe()
     .expect("current executable path invalid")
     .with_file_name("flowistry-driver");
@@ -23,41 +26,20 @@ fn main() {
     (version: VERSION)
     (author: "Will Crichton <wcrichto@cs.stanford.edu>")
     (@setting TrailingVarArg)
+    (@arg BENCH: -b --bench)
     (@subcommand rustc_version =>)
-    (@subcommand backward_slice =>
+    (@subcommand preload =>)
+    (@subcommand spans =>
       (@arg file:)
-      (@arg start:)
-      (@arg end:)
-      (@arg context_mode: --contextmode +takes_value)
-      (@arg mutability_mode: --mutabilitymode +takes_value)
-      (@arg pointer_mode: --pointermode +takes_value)
-      (@arg flags: ...)
-    )
-    (@subcommand forward_slice =>
-      (@arg file:)
-      (@arg start:)
-      (@arg end:)
-      (@arg flags: ...)
-    )
-    (@subcommand find_mutations =>
-      (@arg file:)
-      (@arg start:)
-      (@arg end:)
-      (@arg flags: ...)
-    )
-    (@subcommand effects =>
-      (@arg file:)
-      (@arg pos:)
-      (@arg flags: ...)
-    )
-    (@subcommand decompose =>
-      (@arg file:)
-      (@arg pos:)
       (@arg flags: ...))
     (@subcommand focus =>
       (@arg file:)
       (@arg pos:)
       (@arg flags: ...))
+    (@subcommand decompose =>
+        (@arg file:)
+        (@arg pos:)
+        (@arg flags: ...))
     (@subcommand playground =>
       (@arg file:)
       (@arg start:)
@@ -73,29 +55,17 @@ fn main() {
       println!("{commit_hash}");
       exit(0);
     }
-    ("backward_slice" | "forward_slice" | "playground", Some(sub_m)) => (
-      vec![
-        ("FILE", sub_m.value_of("file").unwrap()),
-        ("START", sub_m.value_of("start").unwrap()),
-        ("END", sub_m.value_of("end").unwrap()),
-        (
-          "CONTEXT_MODE",
-          sub_m.value_of("context_mode").unwrap_or("SigOnly"),
-        ),
-        (
-          "MUTABILITY_MODE",
-          sub_m
-            .value_of("mutability_mode")
-            .unwrap_or("DistinguishMut"),
-        ),
-        (
-          "POINTER_MODE",
-          sub_m.value_of("pointer_mode").unwrap_or("Precise"),
-        ),
-      ],
+    ("preload", _) => {
+      let mut cmd = Command::new(cargo_path);
+      cmd.args(&["check", "--all", "--frozen", "--target-dir", TARGET_DIR]);
+      let exit_status = cmd.status().expect("could not run cargo");
+      exit(exit_status.code().unwrap_or(-1));
+    }
+    ("spans", Some(sub_m)) => (
+      vec![("FILE", sub_m.value_of("file").unwrap())],
       sub_m.value_of("file").unwrap(),
     ),
-    ("find_mutations", Some(sub_m)) => (
+    ("playground", Some(sub_m)) => (
       vec![
         ("FILE", sub_m.value_of("file").unwrap()),
         ("START", sub_m.value_of("start").unwrap()),
@@ -103,7 +73,7 @@ fn main() {
       ],
       sub_m.value_of("file").unwrap(),
     ),
-    ("effects" | "decompose" | "focus", Some(sub_m)) => (
+    ("decompose" | "focus", Some(sub_m)) => (
       vec![
         ("FILE", sub_m.value_of("file").unwrap()),
         ("POS", sub_m.value_of("pos").unwrap()),
@@ -115,10 +85,32 @@ fn main() {
     }
   };
 
+  let sub_m = match matches.subcommand() {
+    (_, Some(sub_m)) => sub_m,
+    _ => unreachable!(),
+  };
+  args.extend([
+    (
+      "CONTEXT_MODE",
+      sub_m.value_of("context_mode").unwrap_or("SigOnly"),
+    ),
+    (
+      "MUTABILITY_MODE",
+      sub_m
+        .value_of("mutability_mode")
+        .unwrap_or("DistinguishMut"),
+    ),
+    (
+      "POINTER_MODE",
+      sub_m.value_of("pointer_mode").unwrap_or("Precise"),
+    ),
+  ]);
+
   let (cmd, flags) = match matches.subcommand() {
     (cmd, Some(sub_m)) => (cmd, sub_m.value_of("flags")),
     _ => unimplemented!(),
   };
+  log::debug!("Command: {cmd}");
   args.push(("COMMAND", cmd));
 
   let file_path = PathBuf::from(file_name);
@@ -127,23 +119,29 @@ fn main() {
     .other_options(["--offline".to_string()])
     .exec()
     .unwrap();
-  let (pkg, target) = metadata
+
+  let workspace_members = metadata
     .workspace_members
     .iter()
-    .filter_map(|pkg_id| {
-      let pkg = metadata
+    .map(|pkg_id| {
+      metadata
         .packages
         .iter()
         .find(|pkg| &pkg.id == pkg_id)
-        .unwrap();
+        .unwrap()
+    })
+    .collect::<Vec<_>>();
 
+  let (pkg, target) = workspace_members
+    .iter()
+    .filter_map(|pkg| {
       let target = pkg
         .targets
         .iter()
         .filter(|target| file_path.starts_with(target.src_path.parent().unwrap()))
-        .max_by_key(|target| target.src_path.components().count());
+        .max_by_key(|target| target.src_path.components().count())?;
 
-      target.map(move |target| (pkg, target))
+      Some((pkg, target))
     })
     .next()
     .unwrap_or_else(|| panic!("Could not find target for path: {file_name}"));
@@ -151,9 +149,20 @@ fn main() {
   let mut cmd = Command::new(cargo_path);
   cmd
     .env("RUSTC_WORKSPACE_WRAPPER", flowistry_rustc_path)
-    .args(&["rustc", "--profile", "check", "-q"]);
+    .args(&[
+      "rustc",
+      "--profile",
+      "check",
+      "--target-dir",
+      TARGET_DIR,
+      "--frozen",
+    ]);
+
+  let bench = matches.is_present("BENCH");
+  cmd.arg(if bench { "-v" } else { "-q" });
 
   // Add compile filter to specify the target corresponding to the given file
+  log::debug!("Package: {}", pkg.name);
   cmd.arg("-p").arg(&pkg.name);
   let kind = &target.kind[0];
   if kind != "proc-macro" {
@@ -168,7 +177,14 @@ fn main() {
 
   // RNG is necessary to avoid caching
   let n = thread_rng().gen::<u64>();
-  cmd.args(&["--", &format!("--flowistry={n}")]).args(flags);
+  cmd.args(&["--", &format!("--flowistry={n}")]);
+
+  // Add args passed from CLI
+  cmd.args(flags);
+
+  // TODO: need to figure out how to download / compile dev-dependencies
+  // // Pass --test to rustc so #[test] functions can be analyzed
+  // cmd.arg("--test");
 
   // FIXME(wcrichto): we should make these CLI args as well, then do
   //   caching on VSCode's side
@@ -176,6 +192,22 @@ fn main() {
     cmd.env(format!("FLOWISTRY_{k}"), v);
   }
 
+  // HACK: if running flowistry on the rustc codebase, this env var needs to exist
+  // for the code to compile
+  if workspace_members.iter().any(|pkg| pkg.name == "rustc-main") {
+    cmd.env("CFG_RELEASE", "");
+  }
+
+  if bench {
+    eprintln!(
+      "{:?}",
+      cmd
+        .get_envs()
+        .map(|(k, v)| vec![k.to_string_lossy(), v.unwrap().to_string_lossy()])
+        .collect::<Vec<_>>()
+    );
+  }
+
   let exit_status = cmd.status().expect("could not run cargo");
   if !exit_status.success() {
     exit(exit_status.code().unwrap_or(-1));
